/*
 * kylin-status-manager
 *
 * Copyright (C) 2022, KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * Authors: wangguixi <wangguixi@kylinos.cn>
 *
 */
#include <QGuiApplication>
#include <QDebug>
#include <QDBusConnection>
#include <QDBusInterface>
#include <QDBusReply>
#include "autoswitch.h"

// 状态管理器
#define SETTINGS_KYLINSTATUSMANAGER    "org.kylin-status-manager"
#define MODE_MONITER         "modemonitor" // 是否监听角度值
#define TABLET_MODE          "tablet-mode"
#define AUTO_ROTATION        "auto-rotation" // 旋转
#define ROTATIONS            "rotations"
// 主题
#define THEME_QT_SCHEMA "org.ukui.style"
#define THEME_Style_Name "styleName"

#define MODE_TYPE       1
#define ROTATIONS_TYPE   0
#define KEYBOARD_TYPE     2

autoSwitch::autoSwitch(QObject *parent) :
    QObject(parent)
{
    qInfo()<<"===initialization===";
    /*日志规则
     * 普通日志qDebug()输出，函数开始和结束用《====****====》加以描述
     * 数据gsetting值等：qInfo()输出
    */
    //注册DBus
    regist_dbus();
    //初始化本地数据gsetting （自动旋转、平板pc、自动切换）
    QByteArray id(SETTINGS_KYLINSTATUSMANAGER);
    if(QGSettings::isSchemaInstalled(id)) {
        status_manager_Setting=new QGSettings(id);
    } else {
        qInfo()<<"org.kylin-status-manager is not install";
        return;
    }

    //主题 （监听主题的gsetting,对外提供dbus接口信号，注：主题没有对外提供此方法，所以由状态管理器提供给搜狗）
    QByteArray id3(THEME_QT_SCHEMA);
    if(QGSettings::isSchemaInstalled(id3))
    {
        ukui_style_Setting = new QGSettings(id3);
        // 监听当前主题变化
        connect(ukui_style_Setting, &QGSettings::changed, [this](const QString & key) {
            qInfo()<<"org.ukui.style changed"<<key;
            if (key == THEME_Style_Name) {
                emit stylename_change_signal(ukui_style_Setting->get(THEME_Style_Name).toString());
            }
        });
    } else {
        qInfo()<<"org.ukui.style is not install";
    }

    //机器类型配置文件
    QString path="/etc/ukui/kylin-status-manager/statusconfig.ini";
    setting=new QSettings(path,QSettings::IniFormat);

    /*设备初始化init()*/

    devicestatus = new Devices_Status_Monitoring;
    init();
    /*设备的状态监控，包括键盘状态和旋转状态，仅仅是获得当前设备的状态情况，
     *不根据设计逻辑进行判断和控制，判断和控制有autoswitch中进行处理
     * 另 该部分初始化应该在机器初始化后进行，防止信号多次发送：
     * connect链接两个信号
     * 两个get方法
   */
    update_keyboard_status();
    connect(devicestatus,&Devices_Status_Monitoring::keyboard_enable_signal,
            this,&autoSwitch::processing_keyboard_enable_signal);
    connect(devicestatus,&Devices_Status_Monitoring::rotations_change_signal,
            this,&autoSwitch::processing_rotations_change_signal);

    //旋转翻折延时防抖处理
    debounce_timer= new QTimer();
    connect(debounce_timer, &QTimer::timeout, this, &autoSwitch::mode_debounce_timeout);
    debounce_timer1= new QTimer();
    connect(debounce_timer1, &QTimer::timeout, this, &autoSwitch::rotations_debounce_timeout);
    debounce_timer2= new QTimer();
    connect(debounce_timer2, &QTimer::timeout, this, &autoSwitch::keyboard_debounce_timeout);

    //该槽函数是为了在通过侧边栏切换时键盘状态不能更新的情况设置的。
    connect(this,&autoSwitch::mode_change_signal,[this](){
        keyboard_debounce(100);
    });
    qInfo()<<"===initialization end===";
}
autoSwitch::~autoSwitch()
{
    if(debounce_timer) {
        delete debounce_timer;
    }
    if(debounce_timer1) {
        delete debounce_timer1;
    }
    if(debounce_timer2) {
        delete debounce_timer2;
    }

    if(ukui_style_Setting) {
        delete ukui_style_Setting;
    }
    if(status_manager_Setting) {
        delete status_manager_Setting;
    }

    delete setting;
    delete devicestatus;
}
void autoSwitch::init()
{
    qDebug()<<__FILE__<<__LINE__<<"====start of initialization====";
    if(status_manager_Setting) {
        ksmstruct.modemonitor = status_manager_Setting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autorotation =status_manager_Setting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tablet_mode=status_manager_Setting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation=status_manager_Setting->get(ROTATIONS).toBool();//旋转状态
    } else {
        qInfo()<<"init fail";
        return;
    }

    //初始化 是否开始自动切换
    if (ksmstruct.modemonitor == true) {
        //根据键盘情况更新模式
        int keyboardenable=get_inputmethod_mode();
        if (keyboardenable == 1) {
            //pc模式
            set_tabletmode(false,"ksm","init");
            set_rotation("normal","ksm","init");
            ksmstruct.keyboardenable=true;
        } else if (keyboardenable == 0) {
            set_tabletmode(true,"ksm","init");
            ksmstruct.keyboardenable=false;
            //根据当前旋转情况初始化
            if(ksmstruct.autorotation == true) {
                QString rotationstatus=devicestatus->get_rotation_status();
                set_rotation(rotationstatus,"ksm","init");
            } else {//没有开全自动旋转是根据上一次的记录还是根据默认情况设置normal??
                set_rotation("normal","ksm","init");//根据默认设置
    //                set_rotation("ksmstruct.rotation","ksm","init");//根据上一次设置
            }
        } else {
            qInfo()<<"初始化获取失败";
            return ;
        }

    } else {
        //恢复上次状态
        set_tabletmode(ksmstruct.tablet_mode,"ksm","init");
        if(ksmstruct.tablet_mode == false) {
            set_rotation("normal","ksm","init");
        } else {//平板
            if(ksmstruct.autorotation == true) {
                QString rotationstatus=devicestatus->get_rotation_status();
                set_rotation(rotationstatus,"ksm","init");
            } else {//没有开全自动旋转是根据上一次的记录还是根据默认情况设置normal??
                set_rotation("normal","ksm","init");//根据默认设置
//                set_rotation("ksmstruct.rotation","ksm","init");//根据上一次设置
            }
        }
    }
    qDebug()<<__FILE__<<__LINE__<<"====End of initialization====";
}
//注册dbus
void autoSwitch::regist_dbus()
{
    qDebug()<<"====RegistDbus====";
    //被拉起后的回调
    QDBusInterface Sessioninterface("org.gonme.SessionManager",
                             "/org/gonme/SessionManager",
                             "org.gonme.SessionManager",
                                QDBusConnection::sessionBus());

    QDBusReply<QString> reply = Sessioninterface.call("startupfinished","kylin-status-manager","startupfinished");

    QDBusConnection  sessionBus= QDBusConnection::sessionBus();
    if (sessionBus.registerService("com.kylin.statusmanager.interface")) {
        qDebug() <<"Regist Dbus Service success";
        sessionBus.registerObject("/", this, QDBusConnection :: ExportAllSlots | QDBusConnection :: ExportAllSignals);
    } else {
        qCritical() << "Register DBus Service Error:" << sessionBus.lastError().message();
    }
    qDebug()<<"====End of RegistDbus====";
}
// 处理系列
void autoSwitch::processing_keyboard_enable_signal(bool enable)
{
    qInfo()<<"processing_keyboard_enable_signal"<<enable;
    if(ksmstruct.modemonitor == true) {//开启自动切换
        if(enable) {
            //当前键盘状态可用
            if(ksmstruct.tablet_mode==true) {//pc模式：模式不一致，改变
                qDebug()<<"模式：模式不一致，改变";
                mode_debounce(100);
            }else {//模式：模式一样，不管
                qDebug()<<"模式：模式一致，不改变";
            }

        } else {
            //当前键盘状态不可用
            if(ksmstruct.tablet_mode==false) {//模式：模式不一致，改变
                qDebug()<<"模式：模式不一致，改变";
                mode_debounce(100);
            }else {//模式：模式一样，不管
                qDebug()<<"模式：模式一致，不改变";
            }
        }
    }else {//未开启自动切换
        //模式切换不变
        qDebug()<<"自动切换未开启";
    }
    keyboard_debounce(100);

}
void autoSwitch::processing_rotations_change_signal(QString rotations)
{
    qInfo()<<"processing_rotations_change_signal"<<rotations;
    //当前旋转与新信号的旋转是否一致
    if(ksmstruct.rotation == rotations) {
        return ;
    }
    //是否是pc模式 (因为自动旋转已经在旋转监控哪进行了处理，所以此处不用在判断自动旋转的开关)
    if(ksmstruct.tablet_mode == true) {
        //防抖缓冲
        rotations_debounce(100);
    }
}
void autoSwitch::mode_debounce_timeout() //模式切换防抖
{
    ++foldtime;
    if (foldtime == 1) {
        foldtime = 0;
        debounce_timer->stop();
        //得到最新的状态进行设置
        update_mode_status();
    }
}
void autoSwitch::rotations_debounce_timeout() //旋转防抖
{
    ++rotetime;
    if (rotetime == 7) {
        rotetime = 0;
        debounce_timer1->stop();
        //得到最新的状态进行设置
        update_rotation_status();
    }
}
void autoSwitch::keyboard_debounce_timeout() //键盘缓冲
{
    ++keyboardtim;
    if (keyboardtim == 16) {
        keyboardtim = 0;
        debounce_timer2->stop();
        //得到最新的状态进行设置
        update_keyboard_status();
    }
}
void autoSwitch::mode_debounce(int msec)
{
    if(debounce_timer->isActive()){
        debounce_timer->stop();
        foldtime=0;
        debounce_timer->start(msec);
    }else{
        debounce_timer->start(msec);
    }
}
void autoSwitch::rotations_debounce(int msec)
{
    if(debounce_timer1->isActive()){
        debounce_timer1->stop();
        keyboardtim=0;
        debounce_timer1->start(msec);
    }else{
        debounce_timer1->start(msec);
    }
}
void autoSwitch::keyboard_debounce(int msec)
{
    if(debounce_timer2->isActive()){
        debounce_timer2->stop();
        rotetime=0;
        debounce_timer2->start(msec);
    }else{
        debounce_timer2->start(msec);
    }
}
//update 系列  不考虑判断，由其他函数判断完成后，调用
//根据当前键盘状态 更新当前键盘情况  不考虑判断
void autoSwitch::update_keyboard_status()
{
    int keyboardenable= devicestatus->get_keyboard_enable_result();
    if(keyboardenable == 1) {
        //键盘可用  当前为物理键盘
        ksmstruct.keyboardenable=true;
        Q_EMIT inputmethod_change_signal(true);
    } else if (keyboardenable == 0) {
        //键盘不可用 当前是虚拟键盘
        ksmstruct.keyboardenable=false;
        Q_EMIT inputmethod_change_signal(false);
    } else {
        qDebug()<<"获取键盘状态失败";
    }
}
//根据当前的gsetting更新缓存结构体数据
void autoSwitch::update_ksmstruct()
{
    if(status_manager_Setting) {
        ksmstruct.modemonitor = status_manager_Setting->get(MODE_MONITER).toBool();//模式监控开关
        ksmstruct.autorotation = status_manager_Setting->get(AUTO_ROTATION).toBool();//旋转监控开关
        ksmstruct.tablet_mode = status_manager_Setting->get(TABLET_MODE).toBool();//机器模式状态
        ksmstruct.rotation = status_manager_Setting->get(ROTATIONS).toBool();//旋转状态
    }

}
//根据当前设备状态 更新模式状态，不管判断条件
void autoSwitch::update_mode_status(){
    qDebug()<<"updateModeStatus";
    int keyboardenable=get_inputmethod_mode();
    if (keyboardenable == 1) {
        //pc模式
        set_tabletmode(false,"ksm","update");
        set_rotation("normal","ksm","update");
    } else if (keyboardenable == 0) {
        set_tabletmode(true,"ksm","update");
        if(ksmstruct.autorotation){
            rotations_debounce(400);

        }
    } else {
        qInfo()<<"获取失败";
        return ;
    }
//    func_to_debounce(KEYBOARD_TYPE);
}
//根据当前最新的旋转状态，获取旋转情况，不考虑判断
void autoSwitch::update_rotation_status()
{
    QString rotationstatus=devicestatus->get_rotation_status();
    set_rotation(rotationstatus,"ksm","update");
}
//get 系列
int autoSwitch::get_inputmethod_mode()//获得当前键盘状态情况
{ 
    return devicestatus->get_keyboard_enable_result();
}
QString autoSwitch::get_current_stylename()//获取当前主体名称 //不用改
{
    if(ukui_style_Setting) {
        return ukui_style_Setting->get(THEME_Style_Name).toString();
    }
}
bool autoSwitch::get_current_tabletmode()//获得当前的模式
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(TABLET_MODE).toBool();
    }
}
QString autoSwitch::get_current_rotation()//获得当前旋转情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(ROTATIONS).toString();
    }
}
bool autoSwitch::get_modemonitor()//获取自动切换模式开关情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(MODE_MONITER).toBool();
    }
}
bool autoSwitch::get_auto_rotation()//获取自动旋转开关情况
{
    if(status_manager_Setting) {
        return status_manager_Setting->get(AUTO_ROTATION).toBool();
    }
}
/*设置接口
 * 修改采用gsetting的方式进行设置，gsetting文件仅仅用来记录
 * 通过函数调用的方式进行更改状态，增加函数复用
*/
bool autoSwitch::set_rotation(QString rotation,QString who,QString why)
{
    qInfo()<<who<<"set_rotation = "<<rotation<<"to"<<why;
    if(rotation!="normal"&&rotation!="upside-down"&&rotation!="left"&& rotation!="right")
    {
        return 0;
    }
    if(status_manager_Setting) {
      status_manager_Setting->set(ROTATIONS,rotation);
    } else {
        return 0;
    }
    if(status_manager_Setting->get(ROTATIONS).toString()==rotation){
        Q_EMIT rotations_change_signal(rotation);
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }
}
bool autoSwitch::set_tabletmode(bool tablet_mode,QString who,QString why)
{
    qInfo()<<who<<"set_tabletmode = "<<tablet_mode<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(TABLET_MODE,tablet_mode);
    } else {
        return 0;
    }
    if(status_manager_Setting->get(TABLET_MODE).toBool()==tablet_mode){
        Q_EMIT mode_change_signal(tablet_mode);
        if(!tablet_mode)//如果切回pc，则状态恢复默认normal
        {
            if(ksmstruct.rotation!="normal")
            {
                status_manager_Setting->set(ROTATIONS,"normal");
                Q_EMIT rotations_change_signal("normal");
            }
        }
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }

}
bool autoSwitch::set_modemonitor(bool modemonitor,QString who,QString why)
{
    qInfo()<<who<<"set_modemonitor = "<<modemonitor<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(MODE_MONITER,modemonitor);
    }else {
        return 0;
    }
    if(status_manager_Setting->get(MODE_MONITER).toBool()==modemonitor){
        Q_EMIT modemonitor_change_signal(modemonitor);
        if(modemonitor) {            
            rotations_debounce(100);
        }
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }
}
bool autoSwitch::set_auto_rotation(bool autorotation,QString who,QString why)
{
    qInfo()<<who<<"set_auto_rotation = "<<autorotation<<"to"<<why;
    if(status_manager_Setting) {
        status_manager_Setting->set(AUTO_ROTATION,autorotation);
    }else {
        return 0;
    }
    if(status_manager_Setting->get(AUTO_ROTATION).toBool()==autorotation){
        Q_EMIT auto_rotation_change_signal(autorotation);
        if(autorotation) {
            update_rotation_status();
        }
        update_ksmstruct();
        return 1;
    }else {
        return 0;
    }
}
QString autoSwitch::get_current_machine_category()
{
    setting->beginGroup("machinecategory");
    QString machine_category= setting->value("machine_category").toString();
    setting->endGroup();
    setting->sync();
    return machine_category;
}

